/*
 *      SenSyn - Version 1.0
 *
 *      Copyright (c) 1991, 1992 by Sensimetrics Corporation
 *      All Rights Reserved.
 *
 */

/*
 *  voice.c - synthesize one voice sample
 *
 *  coded by Eric P. Carlson
 *
 *  Modification History:
 *     7 Nov 1991	removed from sample.c
 *    21 Nov 1991	made spectral_tilt, formant_1s pitch-synchronous
 *    10 Mar 1992	return spectral_tilt(0) when F0 = 0
 */

#include <math.h>		/* needed for power function */
#include "reson.h"		/* resonator support */
#include "synth.h"		/* synthesizer support */

/* Spectral Tilt Bandwidth table */
static int tilt[42] =
  { 5000, 4350, 3790, 3330, 2930, 2700, 2580, 2468, 2364, 2260,
    2157, 2045, 1925, 1806, 1687, 1568, 1449, 1350, 1272, 1199,
    1133, 1071, 1009,  947,  885,  833,  781,  729,  677,  625,
     599,  573,  547,  521,  495,  469,  442,  416,  390,  364,
     338,  312 };

/* LF Voicing Model tables */
static Float bw_lf[41] =
  {   0.0,  -0.6,  -2.0,  -4.0,  -6.0,  -8.0, -10.4, -12.7, -15.3, -17.8,
    -20.1, -22.4, -24.7, -27.0, -29.2, -31.4, -33.6, -35.8, -37.9, -40.0,
    -42.1, -44.1, -46.2, -48.3, -50.4, -52.4, -54.5, -56.6, -57.8, -60.8,
    -62.7, -64.5, -66.3, -68.1, -69.9, -71.6, -73.3, -75.0, -76.6, -78.2,
    -79.6 };

static Float e0_lf[41] =
  { 27.4, 26.3, 25.3, 24.3, 23.2, 22.1, 21.0, 20.0, 18.8, 17.6,
    16.1, 14.9, 13.8, 12.8, 11.7, 10.6, 9.81, 9.00, 8.12, 7.36,
    6.60, 6.05, 5.46, 4.92, 4.41, 3.94, 3.58, 3.14, 2.83, 2.49,
    2.24, 2.03, 1.83, 1.63, 1.48, 1.32, 1.19, 1.08, .982, .902,
    .832 };


/* ---------------------------------------------------------------------- */

Float next_voice_sample(void)
{
  register Float output;
  register Float freq, bw, temp, time, open, seconds, foo;
  register int pulse;
  

  synth.global_time++;	/* flutter time runs continuously */
  
/* check for F0 = zero */
  if (!pars[P_F0]) {
    synth.period_ctr = 0;	/* make sure we open glottis when F0 != 0 */
    synth.glottis_open = 0;
    synth.voicing_state = 0;
    return AdvanceResonator(&spectral_tilt, 0.);
  }

/* glottal opening/closing transition */
  if ((--synth.period_ctr) <= 0) {	/* check for glottal transition */
    if (!synth.glottis_open) {		/* glottis opening */
      synth.glottis_open = 1;
      synth.pulse = 1;
      synth.voicing_time = 0;
      synth.F0 = pars[P_F0];
      synth.FL = pars[P_FL];
      synth.OQ = pars[P_OQ];
      synth.SQ = pars[P_SQ];
      synth.DI = pars[P_DI];
      synth.AV = pars[P_AV];
      synth.TL = pars[P_TL];
      synth.voicing_amp = synth.AV ? dB2amp(spkrdef[S_GV]+synth.AV+A_AV) : 0;
      if (synth.FL) {
        seconds = (float) synth.global_time / spkrdef[S_SR];
        freq = synth.F0 / 10. + (synth.FL / 50. * synth.F0 / 600. *
   		                 (cos(TWO_PI * 12.7 * seconds) +
		                  cos(TWO_PI * 7.1 * seconds) +
		                  cos(TWO_PI * 4.7 * seconds)));
      }
      else
	freq = synth.F0 / 10.;
      synth.pulse_freq = round(spkrdef[S_SR] / freq);
      synth.period_ctr = round(synth.pulse_freq * synth.OQ / 100.);
      synth.close_time = synth.pulse_freq - synth.period_ctr;
      if (synth.DI)
	if (!synth.close_shortened) {
          synth.close_shortened = 1;
          synth.close_time -= round(synth.close_time * synth.DI / 100.);
          synth.voicing_amp *= 1. - synth.DI / 100.;
          synth.TL += round(synth.DI / 4.25);	/** not quite right **/
        }
        else {  /* previous close shortened */
          synth.close_shortened = 0;
	  synth.close_time += round(synth.close_time * synth.DI / 100.);
        }
      else  /* no diplophonia */
	synth.close_shortened = 0;

      /* setup spectral tilt */
      if (spkrdef[S_SS] == 3)	/* LF source corner rounding */
	  synth.TL += 2;
      SetPolePair(&spectral_tilt, round(0.375 * tilt[synth.TL]),
    	  tilt[synth.TL], spkrdef[S_SR]);
      if (synth.TL > 10)
        spectral_tilt.Coef.A *= 1.0 + (synth.TL-10)*(synth.TL-10) / 1000.;

      /* setup pitch-synchronous formant_1_cascade and formant_1_special */
      InterPolePair(&formant_1_cascade, pars[P_F1] + pars[P_DF1],
	  pars[P_B1] + pars[P_DB1], spkrdef[S_SR]);
      InterPolePair(&formant_1_special, pars[P_F1] + pars[P_DF1],
	  pars[P_B1] + pars[P_DB1], spkrdef[S_SR]);

      switch (spkrdef[S_SS]) {
        case 1:  /* impulsive source */
	       synth.voicing_state = 0;
               temp = (Float) synth.pulse_freq * synth.OQ / 100.;
               SetPolePair(&glottal_pulse, 0, round(10000. / temp),
		       spkrdef[S_SR]);
               glottal_pulse.Coef.A *= 0.002675 * temp * temp; /* -51.3 dB */
 	       break;

      case 2:  /* natural source */
               /** nothing to do **/
	       break;

      case 3:  /* LF source */
	       freq = synth.F0 / (22. * synth.OQ / 100.) *
		      (synth.SQ + 100.) / synth.SQ;
               temp = (bw_lf[synth.SQ / 10 - 9] -
		       bw_lf[synth.SQ / 10 - 10]) / 10.;
               bw = (bw_lf[synth.SQ / 10 - 10] + (synth.SQ % 10) * temp) *
		       200. / (synth.pulse_freq * synth.OQ / 100.);
               SetPolePair(&glottal_pulse, round(freq), round(bw),
		       spkrdef[S_SR]);
               temp = (e0_lf[synth.SQ / 10 - 9] -
		       e0_lf[synth.SQ / 10 - 10]) / 10.;
               glottal_pulse.Coef.A *= (e0_lf[synth.SQ / 10 - 10] +
		                        (synth.SQ % 10) * temp) *
		             (synth.pulse_freq * synth.OQ / 100.) / 200.;
	       break;
      }
    }
    else {		/* glottis closing */
      synth.glottis_open = 0;
      synth.voicing_time++;
      synth.pulse = 0;
      synth.period_ctr = synth.close_time;
      InterPolePair(&formant_1_cascade, pars[P_F1], pars[P_B1],
	      spkrdef[S_SR]);
      InterPolePair(&formant_1_special, pars[P_F1], pars[P_B1],
	      spkrdef[S_SR]);
    }
  }
  else {  /* during open or closed phase */
    synth.voicing_time++;
    synth.pulse = 0;
  }


/* ---------------------------------------------------------------------- */
/* voicing source */

  switch (spkrdef[S_SS]) {
    case 1:	/* impulse source */
      pulse = synth.pulse - synth.voicing_state;
      synth.voicing_state = synth.pulse;
      output = AdvanceResonator(&spectral_tilt,
	                   AdvanceResonator(&glottal_pulse,
				   pulse * synth.voicing_amp));
      break;

    case 2:	/* natural KLGLOT88 source */
      if (synth.glottis_open) {
        time = (Float) synth.voicing_time;
        open = (Float) synth.pulse_freq * synth.OQ / 100.;
        temp = synth.voicing_amp * .00055 *
		(synth.pulse_freq ? 100. / synth.pulse_freq : 0) *
		(50. / synth.OQ) *
		(2 * time - 3 * time * time / open);
        output = AdvanceResonator(&spectral_tilt, temp);
      }
      else
	output = AdvanceResonator(&spectral_tilt, 0.);
      break;

    case 3:	/* LF model source */
      if (synth.glottis_open)
        output = AdvanceResonator(&spectral_tilt,
	                   AdvanceResonator(&glottal_pulse,
				   synth.pulse ?
				     (synth.voicing_amp * 0.0795)
				     : 0.));
      else
        ClearResonator(&glottal_pulse);
	output = AdvanceResonator(&spectral_tilt, 0.);
      break;
  }

  return output;
}
